from antievil.utils._info import ObjectInfo


def test_main():
    info: ObjectInfo = ObjectInfo[int](("hello", 2))
    assert str(info) == "hello=<2>"


def test_only_title():
    info: ObjectInfo = ObjectInfo[int]("hello")
    assert str(info) == "hello"


def test_overflow():
    try:
        ObjectInfo[int](("hello", 1, "whocares"))  # type: ignore
    except ValueError:
        return

    raise AssertionError



from antievil._expect import (
    LengthExpectError,
    NameExpectError,
    TypeExpectError,
)


def test_name():
    error: NameExpectError = NameExpectError(
        ("goodbye", 1),
        "hello",
    )

    assert error.args[0] == "goodbye=<1> expected to have name <hello>"


def test_length():
    input_iterable: list[int] = [1, 2, 3]
    error: LengthExpectError = LengthExpectError(
        input_iterable,
        5,
    )

    assertion_message: str = \
        f"iterable <{input_iterable}> expected to be of" \
        f" length <5>, got length <3>"

    assert error.args[0] == assertion_message


def test_type():
    error = TypeExpectError(
        obj=10,
        expected=(str, "instance"),
        ActualType=int,
    )
    assertion_message: str = \
        f"object <{10}> is expected to be instance of type <{str}>" \
        f": got <{int}> instead"
    assert error.args[0] == assertion_message


def test_type_class():
    error = TypeExpectError(
        obj=int,
        expected=(str, "subclass"),
        ActualType=int,
    )
    assertion_message: str = \
        f"object <{int}> is expected to be subclass of type <{str}>" \
        f": got <{int}> instead"
    assert error.args[0] == assertion_message


def test_type_multiple():
    error = TypeExpectError(
        obj=10,
        expected=[(str, "instance"), (bool, "strict")],
        ActualType=int,
    )
    assertion_message: str = \
        f"object <{10}> is expected to be instance of type <{str}>," \
        f" or to strictly have type <{bool}>" \
        f": got <{int}> instead"
    assert error.args[0] == assertion_message



from typing import Generic, Self, TypeVar

ObjectType = TypeVar("ObjectType")
class ObjectInfo(tuple, Generic[ObjectType]):
    """
    Describes an object and it's title.
    """
    __slots__ = ()
    _MIN_LENGTH: int = 1
    _MAX_LENGTH: int = 2

    # we break inheritance principle by limiting accepting argument from
    # Iterable to tuple[...], but i haven't found any way to specify which
    # types i need within my iterable argument without setting it to tuple
    def __new__(
        cls,
        data: Self | tuple[str, ObjectType] | str = ...,
    ) -> Self:

        final_input: tuple[str, ObjectType | None]

        if isinstance(data, cls):
            return data
        elif isinstance(data, tuple):
            tuple_length: int = len(list(data))

            if not cls._MIN_LENGTH <= tuple_length <= cls._MAX_LENGTH:
                raise ValueError(  # noqa: TRY003
                    f"unexpected length <{tuple_length}>",
                )

            final_input = data

        elif isinstance(data, str):
            final_input = (data, None)
        else:
            raise TypeError

        return super().__new__(cls, final_input)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        if self[1] is None:
            return f"{self[0]}"
        return f"{self[0]}=<{self[1]}>"



import inspect
from enum import Enum
from pathlib import Path
from typing import Any, Iterable, Literal

from antievil.utils import ObjectInfo, never

ExpectedInheritanceLiteral = \
    Literal["strict", "instance", "subclass"]


class ExpectedInheritance(Enum):
    """
    Expected type of inheritance.

    Attributes:
        Strict:
            type(a) is B
        Instance:
            isinstance(a, B)
        Subclass:
            issubclass(A, B)
    """
    Strict = "strict"
    Instance = "instance"
    Subclass = "subclass"


class ExpectError(Exception):
    """
    Some value is expected given some object.

    @abstract
    """


class TypeExpectError(ExpectError):
    """
    Some object should have type or by instance of expected type.

    Args:
        obj:
            Object that failed the expectation.
        expected:
            Tuple or tuples (expected to be an any type of it) of expected
            types and chosen type check.
        ActualType(optional):
            Actual type of the object shown in error message. Defaults to None,
            i.e. no actual type will be shown.
    """
    def __init__(
        self,
        *,
        obj: Any,
        expected:
            tuple[type, ExpectedInheritance | ExpectedInheritanceLiteral]
            | list[
                tuple[type, ExpectedInheritance | ExpectedInheritanceLiteral]
            ],
        ActualType: type | None = None,
    ) -> None:
        message: str = f"object <{obj}> is expected to"

        parsed_expected_list: list[tuple[type, ExpectedInheritance]] = []

        if isinstance(expected, tuple):
            parsed_expected_list.append(
                (expected[0], self._convert_expected_inheritance(expected[1])),
            )
        else:
            parsed_expected_list = [
                (e[0], self._convert_expected_inheritance(e[1]))
                for e in expected
            ]

        is_first_part: bool = True
        for parsed_expected in parsed_expected_list:
            part_prefix: str = " " if is_first_part else " or "

            if is_first_part:
                is_first_part = False

            message += \
                part_prefix \
                + self._get_message_part(obj, parsed_expected) \
                + ","

        # remove last comma
        message = message[:-1]

        if ActualType is not None:
            message += f": got <{ActualType}> instead"

        super().__init__(message)

    def _convert_expected_inheritance(
        self, raw: ExpectedInheritance | ExpectedInheritanceLiteral,
    ) -> ExpectedInheritance:
        if isinstance(raw, str):
            return ExpectedInheritance(raw)
        else:
            return raw

    def _get_message_part(
        self, obj: Any, expected: tuple[type, ExpectedInheritance],
    ) -> str:
        match expected[1]:
            case ExpectedInheritance.Strict:
                self._check_is_regular(obj)
                return f"to strictly have type <{expected[0]}>"
            case ExpectedInheritance.Instance:
                self._check_is_regular(obj)
                return f"be instance of type <{expected[0]}>"
            case ExpectedInheritance.Subclass:
                self._check_is_class(obj)
                return  f"be subclass of type <{expected[0]}>"
            case _:
                never(expected[1])
                return None

    def _check_is_class(
        self,
        obj: Any,
    ) -> None:
        if not inspect.isclass(obj):
            raise TypeError

    def _check_is_regular(
        self,
        obj: Any,
    ) -> None:
        if inspect.isclass(obj):
            raise TypeError


class DirectoryExpectError(ExpectError):
    """
    When some path is expected to lead to directory.
    """
    def __init__(
        self,
        *,
        path: Path,
    ) -> None:
        message: str = f"path <{path}> expected to be directory"
        super().__init__(message)


class FileExpectError(ExpectError):
    """
    When some path is expected to lead to non-directory (plain file).
    """
    def __init__(
        self,
        *,
        path: Path,
    ) -> None:
        message: str = f"path <{path}> shouldn't be directory"
        super().__init__(message)


class NameExpectError(ExpectError):
    """
    Some literal name of an object is expected.
    """
    def __init__(
        self,
        objinfo: ObjectInfo | tuple[str, Any] | str,
        name: str,
    ) -> None:
        super().__init__(
            f"{ObjectInfo(objinfo)} expected to have name <{name}>",
        )


class LengthExpectError(ExpectError):
    """
    Some length of an iterable is expected.

    Args:
        iterable:
            Iterable that violated a length expectation.
        expected_length:
            Length expected.
        actual_length(optional):
            Actual length received. If None, it will be calculated out of the
            given iterable. Defaults to None.
    """
    def __init__(
        self,
        iterable: Iterable[Any],
        expected_length: int,
        actual_length: int | None = None,
    ) -> None:
        if actual_length is None:
            actual_length = len(list(iterable))

        super().__init__(
            f"iterable <{iterable}> expected to be of"
            f" length <{expected_length}>, got length <{actual_length}>",
        )



"""
Additional utils for error handling.
"""
from typing import Any, Never, NoReturn

from antievil.utils._info import ObjectInfo


def stringify(value: dict, separator: str = ", ") -> str:
    """
    Transforms dictionary into string representation.
    """
    option_strs: list[str] = []

    for k, v in value.items():
        option_strs.append(f"{k}={v}")

    return separator.join(option_strs)


def get_titled_value(
    title: str,
    value: Any | None = None,
) -> str:
    titled_value: str = title
    if value:
        titled_value = f"{title} <{value!s}>"

    return titled_value


def never(_: Never) -> NoReturn:
    error_message: str = "unhandled case"
    raise NotImplementedError(error_message)



import importlib.metadata

from antievil._expect import (
    DirectoryExpectError,
    ExpectError,
    FileExpectError,
    LengthExpectError,
    NameExpectError,
    TypeExpectError,
)
from antievil._main import (
    AbstractUsageError,
    AlreadyEventError,
    AuthError,
    CannotBeNoneError,
    DisabledAccessTokenError,
    DuplicateNameError,
    EmptyInputError,
    ExpiredTokenError,
    FinalStatusError,
    ForbiddenResourceError,
    IncorrectModelCompositionError,
    LockError,
    LogicError,
    MalformedHeaderAuthError,
    ModeFeatureError,
    NotFoundError,
    NoWebsocketConnectionError,
    OneObjectExpectedError,
    PleaseDefineError,
    RequestError,
    RequiredClassAttributeError,
    StatusChangeError,
    TypeConversionError,
    UnauthorizedError,
    UnmatchedZipComposition,
    UnsetValueError,
    UnsupportedError,
    WrongGenericTypeError,
    WrongPasswordError,
    WrongUsernameError,
)
from antievil.utils import ObjectInfo

__version__ = importlib.metadata.version("antievil")
